#! /usr/bin/env python
# -*- coding: utf-8 -*-
""" Furtive - File Integrity Verification System """

import os
import logging

from furtive.hasher import HashDirectory
from furtive.manifest import Manifest

__version__ = '0.4.0'


class Furtive(object):
    """ Furtive is an application which stores file state and allows users to
        verify the state in the future. Example use cases include file archives
        and file transport.

        If the manifest file exists, it will be automatcally loaded. Calling
        create() will overwrite the existing manifest in memory as well as the
        file.

        :param base_dir: Base directory to use for the manifest. Can be a full
                         or relative path.
        :type base_dir: str
        :param manifest_path: Path to the manifest file. Can be a full or
                              relative path.
        :type manifest_path: str
        :param exclude: list containing patterns to use to exclude files from
                        the manifest.
        :type exclude: list
    """

    def __init__(self, base_dir, manifest_path, exclude=None):
        self.base_dir = base_dir
        self.manifest_path = manifest_path
        self.exclude = [] if exclude is None else exclude
        self.manifest = Manifest(self.base_dir,
                                 self.manifest_path,
                                 self.exclude)

        if os.path.exists(manifest_path):
            self.manifest.load()

    def create(self):
        """ Create and save a new manifest.

            The contents of the new Manfiest() will be saved to
            `manifest_path`.

            :return: None
        """

        self.manifest.create()
        self.manifest.save()

    def compare(self):
        """ Compare the hashes in the database with the current hashes of files
            on the file system.

            :return: Dictionary of added, deleted, and changed files.
            :return type: dict
        """

        if self.manifest.is_empty():
            raise RuntimeError('Manifest {0!s} not loaded.'
                               .format(self.manifest_path))

        logging.info('Generating temporary updated manifest.')
        current_manifest = Manifest(self.base_dir, '/dev/null')
        current_manifest.create()

        logging.info('Generating list of changed files')

        added_files = list(set(current_manifest.manifest) -
                           set(self.manifest.manifest))
        removed_files = list(set(self.manifest.manifest) -
                             set(current_manifest.manifest))
        common_files = list(set(self.manifest.manifest) &
                            set(current_manifest.manifest))
        changed_files = [common_file for common_file in common_files
                         if current_manifest[common_file] !=
                         self.manifest[common_file]]

        return {'removed': removed_files,
                'added': added_files,
                'changed': changed_files}



#!/usr/bin/env python
""" furtive command line tool """


import os
import sys
import logging
import argparse

from argparse import RawTextHelpFormatter

import yaml

from furtive import Furtive, __version__


def parse_args(args=None):
    """ Method to parse command line arguments.

        :param args: command line arguments as typed on the command line
        :type args: str

        :return: parsed arguments from argparse()
        :rtype: argparse.Namespace()
    """

    action_help = ''' Which action to perform:
      compare - compare the current state of the files on the file system with
                the recorded state in the manifest file. Status code is 0 if
                the comparison was successful.
      check   - check the integrity of files listed in the manifest. Same as
                compare but exits with status code 1 if there are changes to
                the files included in the manifest. That is, if any file hash
                changes or if files are added or removed, the application will
                exit with a status code of 1 to indicate there are changes.
                This action can be useful for scripting. For example, to run a
                nightly cron check of a manifest.
      create  - create a new manifest from the files inthe directory
                specified by the --basedir argument.
    '''

    parser = argparse.ArgumentParser(description='Manage a Furtive manifest',
                                     formatter_class=RawTextHelpFormatter)
    parser.add_argument('--basedir', action='store', default='.',
                        help='Directory containing files that will be\n'
                             'checked. Default: .')
    parser.add_argument('--manifest', action='store', dest='manifest_path',
                        default=None,
                        help='Location of the manifest file. Manifests may\n'
                             'be located outside the directory indicated by\n'
                             '--basedir. Must provide path and filename of\n'
                             'the manifest file.\n'
                             'Default: <basedir>/.manifest.yaml')
    parser.add_argument('--log-level', action='store', dest='log_level',
                        choices=('debug', 'info', 'warn', 'error', 'critical'),
                        default='info', help='verbosity of furtive')
    parser.add_argument('--exclude', dest='exclude', action='append',
                        default=[], metavar='PATTERN',
                        help='Patterns to exclude files and directories from\n'
                             'manifest. Can have multiple occurances of this\n'
                             'argument. Excludes are not stored in the\n'
                             'manifest so it is up to the user to provide\n'
                             'the same arguments every run. Patterns are\n'
                             'evaluated as UNIX shell-style wildcard\n'
                             'characters. See the [fnmatch documentation]'
                             '(https://docs.python.org/2/library/fnmatch.html)'
                             'for more information.\n\nIt is important to\n'
                             'note that exclusions are not stored. Therefore\n'
                             ',they must be specified for every run of\n'
                             '`furtive`. Otherwise, the files which were\n'
                             'previously excluded will be included and will\n'
                             'show up as files added to the manifest.')
    parser.add_argument('--quiet', action='store_true',
                        help='Only print out critial error messages. Do not\n'
                        'print a report at the end of a compare run. Using\n'
                        'this argument will override the log-level and set\n'
                        'it to "critical". Only acceptions will be printed\n'
                        'to terminal. The return code will be the only way\n'
                        'to know if a Manifest has changed. This is useful\n'
                        'for scripting such as a cron based manifest checks.\n'
                        'Useful with the check command.')
    parser.add_argument('--report-output', metavar='FILE_NAME',
                        type=argparse.FileType('w'), default='-',
                        help='File to print the diff report to. - for stdout.'
                             'This can be consumed by other scripts to'
                             'determine exactly what has changed within the'
                             'manifest'
                             'Default: -')
    parser.add_argument('--version', action='version',
                        version='%(prog)s {version}'
                        .format(version=__version__))
    parser.add_argument('action', choices=('create', 'compare', 'check'),
                        help=action_help,
                        nargs=1)

    if args is None:
        return parser

    parsed_args = parser.parse_args(args)
    parsed_args.action = parsed_args.action.pop()
    if parsed_args.manifest_path is None:
        parsed_args.manifest_path = os.path.join(parsed_args.basedir,
                                                 '.manifest.yaml')
    if parsed_args.quiet:
        parsed_args.log_level = 'critical'

    return parsed_args


def main():
    """ Starting point of the furtive command line tool """

    args = parse_args(sys.argv[1:])

    log_format = '%(asctime)s [%(levelname)s]: %(message)s'
    date_format = '%c %Z'
    numeric_level = getattr(logging, args.log_level.upper(), None)
    logging.basicConfig(level=numeric_level,
                        format=log_format,
                        datefmt=date_format)

    furtive = Furtive(args.basedir, args.manifest_path, args.exclude)

    if args.action == 'create':
        furtive.create()
    elif args.action in ['compare', 'check']:
        changes = furtive.compare()
        if not args.quiet:
            args.report_output.write(yaml.safe_dump(changes,
                                                    default_flow_style=False))
        if args.action == 'check' and (changes['changed'] or
                                       changes['removed'] or
                                       changes['added']):
            sys.exit(1)



#!/usr/bin/env python

from distutils.core import setup

__VERSION__ = '0.4.0'

long_description = """
Github: https://github.com/dbryant4/furtive
"""

setup(name='Furtive',
      version=__VERSION__,
      description='File Integrity Verification System',
      author='Derrick Bryant',
      author_email='dbryant4@gmail.com',
      long_description=long_description,
      license='MIT',
      packages=['furtive'],
      entry_points={
        'console_scripts': [
            'furtive=furtive.cli:main',
        ]
      },
      url='https://furtive.readthedocs.org',
      download_url='https://github.com/dbryant4/furtive',
      install_requires=[
        'PyYAML==3.11',
        'argparse==1.4.0',
        'future==0.15.2'
      ],
      classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
      ],
     )



#! /usr/bin/env python
# -*- coding: utf-8 -*-

""" Manages the hashing of files """

import os
import hashlib
import logging
import fnmatch
import multiprocessing


def hash_task(file_path, hash_algorithm='md5'):
    """ Responsible for hashing a file.

        This function reads in the file ``file_path`` in small chuncks the size
        of the hash algorithm's block size in order to avoid running out of
        memory. This means that this function should be able to read any file
        irregardless of the size.

        :param file_path: path of file to hash
        :type file_path: str

        :param hash_algorithm: the hashing algorithm to use. All options
                               available in `hashlib.algorithms` should work.
                               See:
                               https://docs.python.org/2/library/hashlib.html
        :type hash_algorithm: str

        :return: hash of file
        :return type: dict
    """

    try:
        if not terminating.is_set():
            with open(file_path, 'rb') as file_to_hash:
                logging.debug('Starting Hash of %s', file_path)
                hash_object = hashlib.new(hash_algorithm)
                while True:
                    chunk = file_to_hash.read(hash_object.block_size)
                    if not chunk:
                        break
                    hash_object.update(chunk)
                file_hash = hash_object.hexdigest()
                logging.debug('Hash for %s: %s', file_path, file_hash)
        else:
            return None

    except KeyboardInterrupt:
        logging.debug('Stopping hash of %s', file_path)
        terminating.set()
        return None

    return {file_path: file_hash}


def initializer(terminating_):
    """ Method to make terminating a global variable so that it is inherited
        by child processes.
    """

    # This places terminating in the global namespace of the worker
    # subprocesses.
    # This allows the worker function to access `terminating` even though it is
    # not passed as an argument to the function.
    global terminating
    terminating = terminating_


class HashDirectory(object):
    """ Object to manage hashing files in a directory.

        This object is responsible for walking the directory tree and
        adding each file to a list. Once the directory walk has compelted, each
        file path is passed to hash_task(). After each file has been
        hashed, this object will then create a Python dictionary of files with
        their associated hash.

        :param directory: Path to directory containing files
        :type directory: str
        :param exclude: list containing patterns to use to exclude files from
                        the manifest.
        :type exclude: list

        :return: Dictionary of file:hash
        :return type: dict
    """

    def __init__(self, directory, exclude=None):
        self.directory = directory
        self.hashes = {}
        self.exclude = [] if exclude is None else exclude

    def hash_files(self):
        """ Orchestrates the discovery and hashing of files.

            Note: This method only supports the md5 hashing algorithm
        """

        files_to_hash = []
        num_processes = multiprocessing.cpu_count() * 2

        logging.info('Discovering files in %s',
                     self.directory)
        for root, _, files in os.walk(self.directory):
            for found_file in files:
                full_path = os.path.join(root, found_file)
                relative_path = os.path.relpath(full_path, self.directory)
                if self.excluded(relative_path):
                    continue
                logging.debug('Found %s', relative_path)
                files_to_hash.append(relative_path)

        self.hashes = {}
        logging.debug('Switching current working directory to %s',
                      self.directory)
        old_cwd = os.getcwd()
        os.chdir(self.directory)
        logging.debug('Starting %s hash worker processes', num_processes)
        terminating = multiprocessing.Event()
        pool = multiprocessing.Pool(initializer=initializer,
                                    initargs=(terminating, ),
                                    processes=num_processes)

        logging.info('Hashing %s files', len(files_to_hash))
        try:
            results = []
            results = pool.map(hash_task, files_to_hash, num_processes*2)
            logging.debug('Stopping hashing processes')
            pool.close()
        except KeyboardInterrupt:
            pool.terminate()
        finally:
            logging.debug('Waiting for processes to stop')
            pool.close()
            pool.join()
            logging.debug('Processes stopped')

        logging.debug('Switching current working directory back to %s',
                      old_cwd)
        os.chdir(old_cwd)

        for item in results:
            self.hashes[list(item.keys())[0]] = list(item.values())[0]

        return self.hashes

    def excluded(self, file_path):
        """ Should the file be excluded from the manifest?

            Determines if a file should be excluded based on UNIX style pattern
            matching. Think *, ?, and [] sequences.

            For matchers, see https://docs.python.org/2/library/fnmatch.html

            :param file_path: path of the file to match against.
            :type file_path: str

            :return: True or False indicating if the file should be
                     excluded from the list of files containted within the
                     manifest.
            :rtype: bool
        """

        for pattern in self.exclude:
            if fnmatch.fnmatchcase(file_path, pattern):
                return True
        return False
